*******************************************************************************;
** SAS macro RemBugNL vs. 0.9,  8 Nov 2002, (C) Jim Groeneveld, Netherlands   *;
*******************************************************************************;
**                                                                            *;
*; %MACRO   RemBugNL   /* REMove BUGgy NewLines in long Clintrial text fields */
/* -------------------------------------------------------------------------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  InFile  =           /* Input  data [drive:][path\]file specification     R */
,  OutFile =           /* Output data [drive:][path\]file specification     R */
,  OverWrit= N         /* specify Y to overwrite an existing OutFile        O */
,  LegalLen= AUTO      /* integer value to force longer lines than present  O */
                       /* AUTO: automatic determination of maximum length     */
,  Version = 0.9       /* forced version specification for version control  O */
,  ForMyUse=           /* argument for the author's use only, do not use it ! */
)/ DES     =             'REMove BUGgy NewLines in text fields'   /* STORE */  ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : REMBUGNL.SAS                                                  *;
%*                                                                            *;
%* Purpose    : REMove BUGgy NewLines in long text fields within Clintrial    *;
%*              export (ascii) files. Replace them by (generally) two spaces  *;
%*              in order to repair original full line oriented records.       *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       :  8 Nov 2002                                                   *;
%* Version    : 0.9 beta                                                      *;
%*                                                                            *;
%* SAS version: 6.12 - 8.2 under Windows 95 - 2K                              *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* FUNCTIONAL SPECIFICATION AND JUSTIFICATION                                 *;
%* ==========================================                                 *;
%*                                                                            *;
%* Summary                                                                    *;
%* -------                                                                    *;
%* Export data files from Clintrial may have embedded Line Breaks, NewLines,  *;
%* Carriage Return / Line Feed characters in long text fields, due to being   *;
%* actually entered while entering long text data in a text field. The macro  *;
%* replaces those NewLines by spaces restoring the maximum line length.       *;
%*                                                                            *;
%* Cause                                                                      *;
%* -----                                                                      *;
%* The problem concerns the presence of illegal, embedded NewLines (CRLFs)    *;
%* within data lines of export files from ClinTrial, which seem to replace    *;
%* other characters (e.g. spaces). This is caused by inserting NewLines into  *;
%* long text fields (to force wrapping) in ClinTrial.                         *;
%* This results in data files with seemingly varying line lengths.            *;
%*                                                                            *;
%* Problem                                                                    *;
%* -------                                                                    *;
%* The problem shows as follows: between the usual data lines with legal      *;
%* contents and full lengths there may be some lines with one or more         *;
%* "embedded" NLs anywhere "within" a line, which seem to end the line,       *;
%* resulting in multiple, shorter lines. Those line parts, including the      *;
%* "embedded" NLs add up to the total line length. These NLs generally seem   *;
%* to occupy the place of 2 other characters (spaces) each. At the last       *;
%* occurrence "in a line" the NL may even replace 0 to 3 fitting spaces.      *;
%*                                                                            *;
%* Solution                                                                   *;
%* --------                                                                   *;
%* This macro program serves as an aid in correcting the problems.            *;
%* It reads an ascii input data file and writes another ascii output data     *;
%* file. It corrects the problem automatically by removing the illegal NLs    *;
%* and inserting spaces for the missing characters, causing all output lines  *;
%* to line up equally (resulting in all equal line lengths).                  *;
%* It may even correct erroneous situations that have not yet been observed.  *;
%*                                                                            *;
%* Strategy                                                                   *;
%* --------                                                                   *;
%* First of all the valid maximum (and minimum) line length is determined by  *;
%* scanning the whole data file in a first pass. This assumes the presence    *;
%* of at least one valid (not cut up) line. If this unfortunately is not the  *;
%* case (which is practically very unlikely) then the LegalLen argument may   *;
%* be specified by the user. The length specified may not be less than the    *;
%* actual maximum line length determined.                                     *;
%*                                                                            *;
%* After reading a line in the second pass the algorithm is as follows:       *;
%* - if a line has the maximum length, i.e. if there are characters (non-     *;
%*   blanks) at the last position (column) then this is OK and gets copied,   *;
%* - if it does not then copy the first part (separate shorter line) without  *;
%*   the NL character, add 2 spaces, add the second and each eventual next    *;
%*   part in the same way, adding 2 spaces each time and append the final     *;
%*   part. Any last part will end the copied line by eventually preceding it  *;
%*   with an exactly fitting number of spaces up to 3. If some part does not  *;
%*   fit at all the line will be padded with spaces and a new one is started. *;
%*   This practically does not occur, but should be checked if it did.        *;
%*   An appropriate warning is issued in such unexpected circumstances.       *;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
%* = internal: - [none]                                                       *;
%* = attached: - [none]                                                       *;
%* = external: - [none]                                                       *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of RemBugNL                                                        *;
%* ===================                                                        *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0 to   15 and                                                            *;
%*     0.2, 16 Apr 1999  initial useful, working versions                     *;
%* 0.3       3 Jun 2002  added deducing max (+ min) line length automatically,*;
%*                         solving the buggy case with a cut up first record, *;
%*                       removed MaxLen argument, being redundant,            *;
%*                       added feedback on line lengths and number of lines,  *;
%*                       added checks on existing input and ouput files       *;
%* 0.9 beta  8 Nov 2002  added file name checks (legal chars in drv\dir\fn),  *;
%*                       added file name macro quoting with special chars,    *;
%*                       added warning message with unexpected circumstances, *;
%*                       solved buggy line count in unexpected circumstances, *;
%*                       added LegalLen arg to force longer lines than found, *;
%*                       added version control and special purpose argument   *;
%*                                                                            *;
%******************************************************************************;

%* OPTIONS NOMPRINT NOMERROR NOSERROR NOMLOGIC NOSYMBOLGEN NOMACROGEN;
%* OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN;
%* remove comment chars for testing ;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN
  %DO;
    %GOTO Finish;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
%* LOCAL macro variables declarations ;
%*____________________________________;

  %LOCAL
   MacName   /* Name of macro: RemBugNL */
   Err       /* text 'E R R O R' as consecutive characters */
   Warn      /* text 'W A R N I N G' as consecutive characters */
   MinLen    /* minimum length of lines (records) of input file */
   MaxLen    /* maximum length of lines (records) of input file */
   Nread     /* number of lines read from input file */
   Nwritten  /* number of lines written to output file */
   Illegal   /* illegal characters in drv:\path\filename */
   Digits    /* legal digits in integer value */
   ;

  %LET MacName = RemBugNL;
  %LET Illegal = %STR(*|%"<>/?); %* Percent char. marks unmatching double quote;
  %LET Digits  = 0123456789; %* legal characters in integer value;
  %LET Err     = %SYSFUNC(COMPRESS(E R R O R, %STR( )));
  %LET Warn    = %SYSFUNC(COMPRESS(W A R N I N G, %STR( )));

%* Macro header in LOG file;
  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |  Macro &MacName, vs. 0.9 by Jim Groeneveld,  8 Nov 2002    |;
  %PUT |     Correction of buggy cut up data lines from ClinTrial    |;
  %PUT |_____________________________________________________________|;
  %PUT;

  %IF (&Version NE AND &Version NE 0.9) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Specified version &Version does not match 0.9,;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;

  %LET Infile  = %SUPERQ(InFile);
  %LET Outfile = %SUPERQ(OutFile);

%* Feedback of specified arguments;
  %PUT &MacName: Input  data file specification : &InFile;
  %PUT &MacName: Output data file specification : &OutFile;
  %PUT &MacName: Overwrite specification        : &OverWrit;
  %PUT &MacName: Legal length specification     : &LegalLen;
  %PUT &MacName: Version nr. specification      : &Version;
  %PUT;
  %PUT &MacName: See SAS log for input and output characteristics;

%* Convert lower case letters to upper case;
  %IF ( &OverWrit NE ) %THEN %LET OverWrit = %UPCASE(%SUBSTR(&OverWrit,1,1));

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Check arguments for contents and discrepancies. ;
*_________________________________________________;

%* InFile;
  %IF (&InFile EQ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Input data file not specified,;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;
      /* checking for all illegal characters (except for control characters) */
  %ELSE %IF ((%LENGTH(&Infile) GT %LENGTH(%QSYSFUNC(COMPRESS(&InFile,&Illegal))))
        /* checking for a maximum of 1 colon in position 2 only */
     OR (%QSUBSTR(&InFile,1,1) EQ : OR %INDEX(%QSUBSTR(&InFile,3),:))) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Input data file name &InFile contains;
    %PUT .                      illegal character(s), macro &MacName will abort;
    %GOTO Finish;
  %END;
  %ELSE %IF ( %SYSFUNC(FILEEXIST(&InFile)) EQ 0 ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Input data file &InFile does not exist,;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;

%* OutFile;
  %IF (&OutFile EQ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Output data file not specified;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;
     /* checking for all illegal characters (except for control characters) */
  %ELSE %IF ((%LENGTH(&Outfile) GT %LENGTH(%QSYSFUNC(COMPRESS(&OutFile,&Illegal))))
        /* checking for a maximum of 1 colon in position 2 only */
     OR (%QSUBSTR(&OutFile,1,1) EQ : OR %INDEX(%QSUBSTR(&OutFile,3),:))) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Output data file name &OutFile contains;
    %PUT .                      illegal character(s), macro &MacName will abort;
    %GOTO Finish;
  %END;
  %ELSE %IF ( %SYSFUNC(FILEEXIST(&OutFile)) EQ 1 AND &OverWrit NE Y ) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Output data file &OutFile exists and;
    %PUT .                       OverWrit is not Y, macro &MacName will abort;
    %GOTO Finish;
  %END;
  %IF( &InFile EQ &OutFile) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: Input and Output data file specifications &InFile;
    %PUT .         are illegally equal, macro &MacName will abort;
    %GOTO Finish;
  %END;

%* LegalLen;
  %IF (&LegalLen EQ OR %UPCASE(&LegalLen) EQ AUTO OR %UPCASE(&LegalLen EQ A))
    %THEN %LET LegalLen = A;
  %ELSE %IF (%LENGTH(%SYSFUNC(COMPRESS(&LegalLen, &Digits))) GT 0) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: LegalLen=&LegalLen contains illegal character(s)
 %SYSFUNC(COMPRESS(&LegalLen, &Digits)),;
    %PUT .                       macro &MacName will abort;
    %GOTO Finish;
  %END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* First data pass: determine maximum line length (main change 3 June 2002) ;
*__________________________________________________________________________;

  DATA _NULL_;
    RETAIN Nread MaxLen 0 MinLen;
    INFILE "&InFile" LENGTH=LineLen LRECL=1048576 END=End;
    INPUT;
    Nread + 1;
    IF (LineLen LT MinLen OR MinLen LT 0) THEN MinLen = LineLen;
    IF (LineLen GT MaxLen) THEN MaxLen = LineLen;
    IF (End) THEN
    DO;
      CALL SYMPUT ('MinLen', COMPRESS(PUT(MinLen, BEST12.)));
      CALL SYMPUT ('MaxLen', COMPRESS(PUT(MaxLen, BEST12.)));
      CALL SYMPUT ('Nread',  COMPRESS(PUT(Nread,  BEST12.)));
    END;
  RUN;
 %PUT &MacName: InFile=&InFile MinLen=&MinLen MaxLen=&MaxLen Nread=&Nread;

  %IF (&LegalLen NE A AND &LegalLen LT &MaxLen) %THEN
  %DO;
    %PUT *** &MacName *** &ERR: LegalLen=&LegalLen is less than actual maximum;
    %PUT .                       length=&MaxLen, macro &MacName will abort;
    %GOTO Finish;
  %END;
  %ELSE %IF (&LegalLen NE A AND &LegalLen GT &MaxLen) %THEN
  %DO;
    %PUT &MacName: Output line length forced to &LegalLen;
    %LET MaxLen = &LegalLen;
  %END;

  %IF (&MinLen EQ &MaxLen) %THEN
  %DO;
    DATA _NULL_;
      INFILE "&InFile" LRECL=&MaxLen;
      FILE  "&OutFile" LRECL=&MaxLen;
      INPUT;
      PUT _INFILE_;
    RUN;
   %PUT &MacName: Input file seems OK, not repaired, only copied to output file;
  %END;

  %ELSE
  %DO;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* Second data pass: start loop reading and writing line contents ;
*________________________________________________________________;

    DATA _NULL_;
      INFILE "&InFile" LRECL=&MaxLen MISSOVER LENGTH=ReadLen END=End;
      FILE  "&OutFile" LRECL=&MaxLen;

      ARRAY C $1 C1 - C&MaxLen;
      RETAIN MaxLen &MaxLen WriteLen 0 Nwritten 0;

* Read lines;
      DO _I_ = 1 TO MaxLen;
        INPUT @_I_ C $CHAR1. @;
      END;
      INPUT; * Finish reading current input line;
      Nread + 1;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* If there is insufficient space left on an output line to append a partial    ;
* input line fill the remaining positions with spaces up to the MaxLen.        ;
* The partial input line is written later below to a new output line.          ;
* This generally is expected to occur only with legal new lines.               ;
*______________________________________________________________________________;

      IF (ReadLen + WriteLen GT MaxLen) THEN LINK PadSpace;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* If a partial line already has been written insert 2 spaces before writing the;
* next part, as those are assumed to substitute the CRLF (2 characters) between;
* words in a long text field. This may, however, be a wrong assumption.        ;
*______________________________________________________________________________;

      IF (WriteLen GT 0 AND WriteLen LT MaxLen) THEN DO;
        DO _I_ = 1 TO 2; * Append (one or) two spaces if possible;
          IF (ReadLen + WriteLen LT MaxLen) THEN DO;
            PUT @(WriteLen+1) ' ' @; * write additional explicit space;
            WriteLen + 1; * Keep up WriteLen;
          END;
        END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* If by subsequent writing the ReadLen part the resulting line becomes short   ;
* for only one character initially an extra additional space has to be written.;
* This adaptation is necessary because the bug shows itself this way as well.  ;
*______________________________________________________________________________;

        IF (ReadLen+WriteLen+1 EQ MaxLen) THEN DO;
          PUT @(WriteLen+1) ' ' @; * write additional explicit space;
          WriteLen + 1; * Keep up WriteLen;
        END;
      END; * << Append space (one or) two (or even three) times >>;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* As long as sum of the part read and the part already written does not exceed ;
* MaxLen append the read part by writing, without closing the line.            ;
*______________________________________________________________________________;

      IF (ReadLen + WriteLen LE MaxLen) THEN DO;
        DO _I_ = 1 TO ReadLen;
          PUT @(_I_+WriteLen) C @;
        END;
        WriteLen = WriteLen + ReadLen;
*       ReadLen = 0; * gets reset by LENGTH= in INFILE statement;
      END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* If exactly a full line has been written then close it (NL) and reset WriteLen;
*______________________________________________________________________________;

      IF (WriteLen EQ MaxLen) THEN DO;
        PUT; * Finish writing current input line;
        Nwritten + 1;
        WriteLen = 0; * reset for new line;
      END;

*~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~;
* If the input file has been read completely, finalize the output file neatly ;
*_____________________________________________________________________________;

      IF (End) THEN
      DO;
        * If Final partial line written, pad with spaces;
        IF (WriteLen) THEN LINK PadSpace;
        CALL SYMPUT ('Nwritten', COMPRESS(PUT(Nwritten, BEST12.)));
      END;

      RETURN;

*~~~~~~~~~~~~~~~~~~~~~~~~~;
* End of repair algorithm ;
*_________________________;

*~~~~~~~~~~~~~~~~~~~~~~~~~;
* Linked routine PadSpace ;
*_________________________;

      PadSpace:
        DO _I_ = WriteLen+1 TO MaxLen; * Pad the line with spaces up to MaxLen;
          PUT @_I_ ' ' @;
        END;
*       WriteLen = MaxLen; * but gets reset below immediately;
        PUT; * Finish writing current input line;
        Nwritten + 1;
        WriteLen = 0; * reset for new line;
%* Use macro %PUT for warns because of PUT to FILE;
        CALL EXECUTE ('%PUT *** &MacName *** &WARN: input line ' ||
                TRIM(COMPRESS(PUT(Nread,BEST12.0))) ||
                ' does not fit at end of;' ||
           '%PUT output line ' || TRIM(COMPRESS(PUT(Nwritten,BEST12.0))) ||
                '. Please check original and result carefully!;');
      RETURN;

*---------;

    RUN;
    %PUT &MacName: Lines read=&Nread, lines written=&Nwritten with length=&MaxLen;
  %END;

  %Finish:
%*************;
%MEND RemBugNL; * -------------------end-of-macro-RemBugNL---------------------;
**************;

